ELF64HEADER = {
    "e_ident": bytearray(16),
    "e_type": bytearray(2),
    "e_machine": bytearray(2),
    "e_version": bytearray(4),
    "e_entry": bytearray(8),
    "e_phoff": bytearray(8),
    "e_shoff": bytearray(8),
    "e_flags": bytearray(4),
    "e_ehsize": bytearray(2),
    "e_phentsize": bytearray(2),
    "e_phnum": bytearray(2),
    "e_shentsize": bytearray(2),
    "e_shnum": bytearray(2),
    "e_shstrndx": bytearray(2)
}

ELF64PROGRAMHEADER = {
    "p_type": bytearray(4),
    "p_offset": bytearray(4),
    "p_vaddr": bytearray(8),
    "p_paddr": bytearray(8),
    "p_filesz": bytearray(8),
    "p_memsz": bytearray(8),
    "p_flags": bytearray(8),
    "p_align": bytearray(8)
}

ELF32HEADER = {
    "e_ident": bytearray(16),
    "e_type": bytearray(2),
    "e_machine": bytearray(2),
    "e_version": bytearray(4),
    "e_entry": bytearray(4),
    "e_phoff": bytearray(4),
    "e_shoff": bytearray(4),
    "e_flags": bytearray(4),
    "e_ehsize": bytearray(2),
    "e_phentsize": bytearray(2),
    "e_phnum": bytearray(2),
    "e_shentsize": bytearray(2),
    "e_shnum": bytearray(2),
    "e_shstrndx": bytearray(2)
}

ELF32PROGRAMHEADER = {
    "p_type": bytearray(4),
    "p_offset": bytearray(4),
    "p_vaddr": bytearray(4),
    "p_paddr": bytearray(4),
    "p_filesz": bytearray(4),
    "p_memsz": bytearray(4),
    "p_flags": bytearray(4),
    "p_align": bytearray(4)
}

ELF32SECTIONHEADER = {
    "sh_name": bytearray(4),
    "sh_type": bytearray(4),
    "sh_flags": bytearray(4),
    "sh_addr": bytearray(4),
    "sh_offset": bytearray(4),
    "sh_size": bytearray(4),
    "sh_link": bytearray(4),
    "sh_info": bytearray(4),
    "sh_addralign": bytearray(4),
    "sh_entsize": bytearray(4)
}

ELF64SECTIONHEADER = {
    "sh_name": bytearray(4),
    "sh_type": bytearray(4),
    "sh_flags": bytearray(8),
    "sh_addr": bytearray(8),
    "sh_offset": bytearray(8),
    "sh_size": bytearray(8),
    "sh_link": bytearray(4),
    "sh_info": bytearray(4),
    "sh_addralign": bytearray(8),
    "sh_entsize": bytearray(8)
}

# Architecture-specific information is tied to the CPU. ELF files of specified
# architectures are expected to contain these values.
ARCH = {
    b"\x03\x00": {
        "e_machine": "386",
        "ei_class": 1,
        "ei_data": 1,
        "ei_version": 1,
        "ei_osabi": 0,
        "ei_abiversion": 0,
        "padding": b"\x00\x00\x00\x00\x00\x00\x00",
        "e_phoff": 0x34,
        "e_ehsize": 0x34,
        "e_phentsize": 0x20,
        "e_shentsize": 0x28,
        "e_version": 1
    },
    b"\x3E\x00": {
        "e_machine": "AMD64",
        "ei_class": 2,
        "ei_data": 1,
        "ei_version": 1,
        "ei_osabi": 0,
        "ei_abiversion": 0,
        "padding": b"\x00\x00\x00\x00\x00\x00\x00",
        "e_phoff": 0x40,
        "e_ehsize": 0x40,
        "e_phentsize": 0x38,
        "e_shentsize": 0x40,
        "e_version": 1
    }
}



import os
import logging
from copy import deepcopy
from struct import unpack, pack

import lepton.utils.constants as constants
import lepton.utils.exceptions as exceptions
import lepton.utils.structures as structures

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
LOG = logging.getLogger(__name__)


class ELFSectionHeaderTable:
    def __init__(self, data, elfheader, new_header):
        """
        :param data: Input file contents
        :type data: <class 'bytes'>
        :param elfheader: Lepton-constructed ELF header contents
        :type elfheader: dict
        :param new_header: Flag to indicate reconstruction of ELF section header
        :type new_header: bool
        """
        self.entries = self.build_section_header_table(data, elfheader, new_header)

    @staticmethod
    def _update_elf32_values(data, sheader_struct, sh_off, shdr_num):
        """
        Update ELF32 section header table entry fields with values.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param sheader_struct: ELF32 section header table entry definition.
        :type sheader_struct: dict
        :param sh_off: Section header table offset
        :type sh_off: int
        :param shdr_num: Section header table entry number. 0-indexed.
        :type shdr_num: int
        :return: Constructed ELF32 section header table entry
        :rtype: dict
        """
        # Calculate offset of s_hdr
        # s_hdr_offset = e_shoff
        e_shentsize = unpack('<H', data[constants.ELF32HEADEROFFSETS.E_SHENTSIZE[0]:
                                        constants.ELF32HEADEROFFSETS.E_SHENTSIZE[1]])[0]
        # s_hdr_i_offset = e_shoff + e_shentsize * shdr_num
        shdr_entry_offset = sh_off + e_shentsize * shdr_num

        shdr = deepcopy(sheader_struct)

        # The section header fields may be inaccurate.
        shdr["sh_name"] = data[shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_NAME[0]:
                               shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_NAME[1]]
        shdr["sh_type"] = data[shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_TYPE[0]:
                               shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_TYPE[1]]
        shdr["sh_flags"] = data[shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_FLAGS[0]:
                                shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_FLAGS[1]]
        shdr["sh_addr"] = data[shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_ADDR[0]:
                               shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_ADDR[1]]
        shdr["sh_offset"] = data[shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_OFFSET[0]:
                                 shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_OFFSET[1]]
        shdr["sh_size"] = data[shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_SIZE[0]:
                               shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_SIZE[1]]
        shdr["sh_link"] = data[shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_LINK[0]:
                               shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_LINK[1]]
        shdr["sh_info"] = data[shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_INFO[0]:
                               shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_INFO[1]]
        shdr["sh_addralign"] = data[shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_ADDRALIGN[0]:
                                    shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_ADDRALIGN[1]]
        shdr["sh_entsize"] = data[shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_ENTSIZE[0]:
                                  shdr_entry_offset + constants.ELF32SECTIONHEADEROFFSETS.SH_ENTSIZE[1]]

        return shdr

    @staticmethod
    def _update_elf64_values(data, sheader_struct, sh_off, shdr_num):
        """
        Update ELF64 section header table entry fields with values.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param sheader_struct: ELF64 section header table entry definition.
        :type sheader_struct: dict
        :param sh_off: Section header table offset
        :type sh_off: int
        :param shdr_num: Section header table entry number. 0-indexed.
        :type shdr_num: int
        :return: Constructed ELF32 section header table entry
        :rtype: dict
        """
        # Calculate offset of s_hdr
        # s_hdr_offset = e_shoff
        e_shentsize = unpack('<H', data[constants.ELF64HEADEROFFSETS.E_SHENTSIZE[0]:
                                        constants.ELF64HEADEROFFSETS.E_SHENTSIZE[1]])[0]
        # s_hdr_i_offset = e_shoff + e_shentsize * shdr_num
        shdr_entry_offset = sh_off + e_shentsize * shdr_num

        shdr = deepcopy(sheader_struct)

        # The section header fields may be inaccurate.
        shdr["sh_name"] = data[shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_NAME[0]:
                               shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_NAME[1]]
        shdr["sh_type"] = data[shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_TYPE[0]:
                               shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_TYPE[1]]
        shdr["sh_flags"] = data[shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_FLAGS[0]:
                                shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_FLAGS[1]]
        shdr["sh_addr"] = data[shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_ADDR[0]:
                               shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_ADDR[1]]
        shdr["sh_offset"] = data[shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_OFFSET[0]:
                                 shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_OFFSET[1]]
        shdr["sh_size"] = data[shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_SIZE[0]:
                               shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_SIZE[1]]
        shdr["sh_link"] = data[shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_LINK[0]:
                               shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_LINK[1]]
        shdr["sh_info"] = data[shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_INFO[0]:
                               shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_INFO[1]]
        shdr["sh_addralign"] = data[shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_ADDRALIGN[0]:
                                    shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_ADDRALIGN[1]]
        shdr["sh_entsize"] = data[shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_ENTSIZE[0]:
                                  shdr_entry_offset + constants.ELF64SECTIONHEADEROFFSETS.SH_ENTSIZE[1]]

        return shdr

    def _build_shdr(self, data, sheader_struct, sh_off, shdr_num):
        """
        Router function. Calls relevant functions to build section header table
        for an ELF32/ELF64 file.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param sheader_struct: ELF32/64 section header table entry definition.
        :type sheader_struct: dict
        :param sh_off: Section header table offset
        :type sh_off: int
        :param shdr_num: Section header table entry number. 0-indexed.
        :type shdr_num: int
        :return: Constructed ELF section header table
        :rtype: list of dict
        """
        if sheader_struct == structures.ELF32SECTIONHEADER:
            return self._update_elf32_values(data, sheader_struct, sh_off, shdr_num)
        elif sheader_struct == structures.ELF64SECTIONHEADER:
            return self._update_elf64_values(data, sheader_struct, sh_off, shdr_num)
        else:
            raise exceptions.ELFSectionHeaderTableError("Invalid Section"
                                                        "Header Structure.")

    @staticmethod
    def _verify_section_header_table(data, shoff_, shentsize, ei_class):
        """
        Using various heuristics, verify if a section header table exists at
        shoff_ offset.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param shoff_: Suspected section headers table offset
        :type shoff_: int
        :param shentsize: Size of each entry in the section header table
        :data shentsize: int
        :param ei_class: Leptop-constructed EI_CLASS field in ELF header
        :type ei_class: int
        """
        # Number of sections should always be an integer which means the size
        # of the section header table must be divisible by the size of each
        # section header table entry. This assumes that the section header
        # table extends all the way to the end of the file.
        filesize = len(data)
        if (filesize - shoff_) % shentsize != 0:
            return False

        # Get a few entries from the section header table.
        # sh_flags field value must be as expected for sh_type == PROGBITS.
        # I'm checking multiple entries (instead of one) to make for a better
        # heuristic.
        match = 0
        check = 5
        current_entry = shoff_
        while check:
            current_entry = current_entry + shentsize
            if ei_class == 1:
                sh_type_start = current_entry + (constants.ELF32SECTIONHEADEROFFSETS.SH_NAME[1] -
                                                 constants.ELF32SECTIONHEADEROFFSETS.SH_NAME[0])
                sh_type_len = constants.ELF32SECTIONHEADEROFFSETS.SH_TYPE[1] -\
                              constants.ELF32SECTIONHEADEROFFSETS.SH_TYPE[0]
                current_entry_sh_type = unpack("<I", data[sh_type_start: sh_type_start + sh_type_len])[0]
                if current_entry_sh_type != 0x1:
                    # Not sh_type == PROGBITS
                    check -= 1
                    continue
                sh_flags_start = sh_type_start + (constants.ELF32SECTIONHEADEROFFSETS.SH_TYPE[1] -
                                                  constants.ELF32SECTIONHEADEROFFSETS.SH_TYPE[0])
                sh_flags_len = constants.ELF32SECTIONHEADEROFFSETS.SH_FLAGS[1] -\
                               constants.ELF32SECTIONHEADEROFFSETS.SH_FLAGS[0]
                current_entry_sh_flags = unpack("<I", data[sh_flags_start: sh_flags_start + sh_flags_len])[0]
            elif ei_class == 2:
                sh_type_start = current_entry + (constants.ELF64SECTIONHEADEROFFSETS.SH_NAME[1] -
                                                 constants.ELF64SECTIONHEADEROFFSETS.SH_NAME[0])
                sh_type_len = constants.ELF64SECTIONHEADEROFFSETS.SH_TYPE[1] - \
                              constants.ELF64SECTIONHEADEROFFSETS.SH_TYPE[0]
                current_entry_sh_type = unpack("<I", data[sh_type_start: sh_type_start + sh_type_len])[0]
                if current_entry_sh_type != 0x1:
                    # Not sh_type == PROGBITS
                    check -= 1
                    continue
                sh_flags_start = sh_type_start + (constants.ELF64SECTIONHEADEROFFSETS.SH_TYPE[1] -
                                                  constants.ELF64SECTIONHEADEROFFSETS.SH_TYPE[0])
                sh_flags_len = constants.ELF64SECTIONHEADEROFFSETS.SH_FLAGS[1] - \
                               constants.ELF64SECTIONHEADEROFFSETS.SH_FLAGS[0]
                current_entry_sh_flags = unpack("<Q", data[sh_flags_start: sh_flags_start + sh_flags_len])[0]
            else:
                raise exceptions.ELFSectionHeaderTableError("Invalid EI_CLASS")
            # When SHF_WRITE and/or SHF_ALLOC and/or SHF_EXECINSTR and/or SHF_MASKPROC
            # are set, they result in some set values: (>0 <=7) or (>=0xF0000000 <=0xF0000007)
            if 0x7 < current_entry_sh_flags < 0xf0000000 or current_entry_sh_flags > 0xf0000007:
                return False
            match += 1
            check -= 1

        return match > 0

    def find_section_header_table(self, data, shentsize, ei_class):
        """
        In stripped ELF files, section header information is removed from the
        ELF header making debugging difficult. This function attempts to find
        the ELF section header in the file contents. Of course, it may not
        always be correct.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param shentsize: Size of each entry in the section header table
        :data shentsize: int
        :param ei_class: Leptop-constructed EI_CLASS field in ELF header
        :type ei_class: int
        """
        shnum = 0
        shoff = 0

        # The initial entry in the section header table is full of null bytes.
        # This is not true in all cases. From the ELF specification:
        # "The initial entry is used in ELF extensions for e_phnum, e_shnum, and
        # e_shstrndx; in other cases, each field in the initial entry is set
        # to zero."
        null_bytes_section = shentsize * b"\x00"

        # Usually, the section header table exists at the end of the file.
        # Again, this may not always be true. Assuming it is true, start
        # search for null_bytes_section from end of file and process from
        # the first found index
        occurrences = []
        current_index = len(data)
        while current_index >= 0:
            current_index = data.rfind(null_bytes_section, 0, current_index)
            if current_index >= 0:
                occurrences.append(current_index)
                current_index -= 1

        for i in occurrences:
            flag = self._verify_section_header_table(data, i, shentsize, ei_class)
            if flag:
                shoff = i
                filesize = len(data)
                shnum = (filesize - shoff) // shentsize
                break

        return shnum, shoff

    @staticmethod
    def find_section_name_string_table(data, sh_num, sh_off,
                                       shentsize, ei_class):
        """
        Determines the index of the section name string table in the sections
        headers table.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param sh_num: Lepton-determined number of entries in section headers
                       table
        :type sh_num: int
        :param sh_off: Lepton-determined section headers table offset
        :type sh_off: int
        :param shentsize: Size of each section headers table entry
        :type shentsize: int
        :param ei_class: Leptop-constructed EI_CLASS field in ELF header
        :type ei_class: int
        """
        for num in range(sh_num):
            current_entry = sh_off + num * shentsize
            if ei_class == 1:
                # ELF32
                sh_type_start = current_entry + (constants.ELF32SECTIONHEADEROFFSETS.SH_NAME[1] -
                                                 constants.ELF32SECTIONHEADEROFFSETS.SH_NAME[0])
                sh_type_len = constants.ELF32SECTIONHEADEROFFSETS.SH_TYPE[1] - \
                              constants.ELF32SECTIONHEADEROFFSETS.SH_TYPE[0]
            elif ei_class == 2:
                sh_type_start = current_entry + (constants.ELF64SECTIONHEADEROFFSETS.SH_NAME[1] -
                                                 constants.ELF64SECTIONHEADEROFFSETS.SH_NAME[0])
                sh_type_len = constants.ELF64SECTIONHEADEROFFSETS.SH_TYPE[1] - \
                              constants.ELF64SECTIONHEADEROFFSETS.SH_TYPE[0]
            else:
                raise exceptions.ELFSectionHeaderTableError(f"Invalid EI_CLASS: {ei_class}")

            current_entry_sh_type = unpack("<I", data[sh_type_start: sh_type_start + sh_type_len])[0]
            if current_entry_sh_type != 3:
                # sh_type != SHT_STRTAB
                continue

            if ei_class == 1:
                # ELF32
                sh_offset_start = current_entry + (constants.ELF32SECTIONHEADEROFFSETS.SH_NAME[1] -
                                                   constants.ELF32SECTIONHEADEROFFSETS.SH_NAME[0]) \
                                                + (constants.ELF32SECTIONHEADEROFFSETS.SH_TYPE[1] -
                                                   constants.ELF32SECTIONHEADEROFFSETS.SH_TYPE[0]) \
                                                + (constants.ELF32SECTIONHEADEROFFSETS.SH_FLAGS[1] -
                                                   constants.ELF32SECTIONHEADEROFFSETS.SH_FLAGS[0]) \
                                                + (constants.ELF32SECTIONHEADEROFFSETS.SH_ADDR[1] -
                                                   constants.ELF32SECTIONHEADEROFFSETS.SH_ADDR[0])
                sh_offset_len = constants.ELF32SECTIONHEADEROFFSETS.SH_OFFSET[1] - \
                                constants.ELF32SECTIONHEADEROFFSETS.SH_OFFSET[0]
                sh_offset = unpack("<I", data[sh_offset_start: sh_offset_start + sh_offset_len])[0]

                sh_size_start = sh_offset_start + sh_offset_len
                sh_size_len = constants.ELF32SECTIONHEADEROFFSETS.SH_SIZE[1] - \
                              constants.ELF32SECTIONHEADEROFFSETS.SH_SIZE[0]
                sh_size = unpack("<I", data[sh_size_start: sh_size_start + sh_size_len])[0]
            elif ei_class == 2:
                sh_offset_start = current_entry + (constants.ELF64SECTIONHEADEROFFSETS.SH_NAME[1] -
                                                   constants.ELF64SECTIONHEADEROFFSETS.SH_NAME[0]) \
                                  + (constants.ELF64SECTIONHEADEROFFSETS.SH_TYPE[1] -
                                     constants.ELF64SECTIONHEADEROFFSETS.SH_TYPE[0]) \
                                  + (constants.ELF64SECTIONHEADEROFFSETS.SH_FLAGS[1] -
                                     constants.ELF64SECTIONHEADEROFFSETS.SH_FLAGS[0]) \
                                  + (constants.ELF64SECTIONHEADEROFFSETS.SH_ADDR[1] -
                                     constants.ELF64SECTIONHEADEROFFSETS.SH_ADDR[0])
                sh_offset_len = constants.ELF64SECTIONHEADEROFFSETS.SH_OFFSET[1] - \
                                constants.ELF64SECTIONHEADEROFFSETS.SH_OFFSET[0]
                sh_offset = unpack("<Q", data[sh_offset_start: sh_offset_start + sh_offset_len])[0]

                sh_size_start = sh_offset_start + sh_offset_len
                sh_size_len = constants.ELF64SECTIONHEADEROFFSETS.SH_SIZE[1] - \
                              constants.ELF64SECTIONHEADEROFFSETS.SH_SIZE[0]
                sh_size = unpack("<Q", data[sh_size_start: sh_size_start + sh_size_len])[0]
            else:
                raise exceptions.ELFSectionHeaderTableError(f"Invalid EI_CLASS: {ei_class}")

            section_data = data[sh_offset: sh_offset + sh_size]
            if b".dynsym" in section_data and b".init" in section_data:
                return num

        return 0

    def build_section_header_table(self, data, elfheader, new_header):
        """
        Builds the section header table of the input ELF file. When building a
        new section header this function relies on Lepton-constructed e_shnum,
        ei_ident fields. It finds a possible e_shoff and e_shnum and constructs
        a section header table based off of that. Values of the section header
        table are read directly from the ELF file based on calculated offsets.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param elfheader: Lepton-constructed ELF header contents
        :type elfheader: dict
        :param new_header: Flag to indicate reconstruction of ELF section header
        :type new_header: bool
        :return: Constructed ELF section header table
        :rtype: list of dict
        """
        shdr_table = []
        sh_num = unpack("<H", elfheader["e_shnum"])[0]
        shstrndx = unpack("<H", elfheader["e_shstrndx"])[0]
        ei_class = elfheader["e_ident"][4]

        if ei_class == 1:
            sh_off = unpack("<I", elfheader["e_shoff"])[0]
        elif ei_class == 2:
            sh_off = unpack("<Q", elfheader["e_shoff"])[0]
        else:
            raise exceptions.ELFSectionHeaderTableError(f"Invalid EI_CLASS: {ei_class}")

        # To reconstruct the ELF section header, all that's needed is to
        # determine sh_off and sh_num fields and updated the ELF header
        # entry
        if new_header:
            shentsize = unpack("<H", elfheader["e_shentsize"])[0]
            sh_num, sh_off = self.find_section_header_table(data, shentsize,
                                                            ei_class)
            shstrndx = self.find_section_name_string_table(data, sh_num, sh_off,
                                                           shentsize, ei_class)

        for shdr_num in range(sh_num):
            if ei_class == 1:
                elfheader["e_shnum"], elfheader["e_shoff"], elfheader["e_shstrndx"] = \
                    pack("<H", sh_num), pack("<I", sh_off), pack("<H", shstrndx)
                shdr = self._build_shdr(data, structures.ELF32SECTIONHEADER,
                                        sh_off, shdr_num)
            else:
                elfheader["e_shnum"], elfheader["e_shoff"], elfheader["e_shstrndx"] = \
                    pack("<H", sh_num), pack("<Q", sh_off), pack("<H", shstrndx)
                shdr = self._build_shdr(data, structures.ELF64SECTIONHEADER,
                                        sh_off, shdr_num)

            shdr_table.append(shdr)

        return shdr_table

    def to_bytes(self):
        """
        Returns ELF section header contents in the form of a sequence of bytes
        :return: ELF section header contents
        :rtype: <class 'bytes'>
        """
        shdr_table_bytes = bytes()
        for shdr in self.entries:
            header_bytes = shdr["sh_name"] + shdr["sh_type"] + \
                           shdr["sh_flags"] + shdr["sh_addr"] + \
                           shdr["sh_offset"] + shdr["sh_size"] + \
                           shdr["sh_link"] + shdr["sh_info"] + \
                           shdr["sh_addralign"] + shdr["sh_entsize"]
            shdr_table_bytes += header_bytes

        return shdr_table_bytes



# TODO: Everything

class Error(Exception):
    """Base class for exceptions in this module"""
    pass


class ELFMagicError(Error):
    """Exception raised if the EI_MAG0 - EI_MAG3 are not 7f 45 4c 46"""

    def __init__(self, message):
        self.message = message


class ELFHeaderError(Error):
    """Exception raised if there is a problem with the ELF header"""

    def __init__(self, message):
        self.message = message


class ELFProgramHeaderTableError(Error):
    """Exception raised if there is a problem with reading or composing the
    program header table"""

    def __init__(self, message):
        self.message = message


class ELFSectionHeaderTableError(Error):
    """Exception raised if there is a problem with reading or composing the
    section header table"""

    def __init__(self, message):
        self.message = message



import os
import logging
from copy import deepcopy
from struct import pack, unpack

import lepton.utils.constants as constants
import lepton.utils.exceptions as exceptions
import lepton.utils.structures as structures

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
LOG = logging.getLogger(__name__)


class ELFHeader:
    def __init__(self, data, new_header):
        """
        :param data: Input file contents
        :type data: <class 'bytes'>
        :param new_header: Flag to decide whether to build a new header with
                           standard values or clone the header of the input file.
        :type new_header: bool
        """
        self.e_machine = None
        self.arch_metadata = None

        if data[constants.MAGIC_SLICE[0]:constants.MAGIC_SLICE[1]] != constants.ELFMAGIC:
            raise exceptions.ELFMagicError("ELFMagicError: Not an ELF: " +
                                           hex(unpack("<I", data[constants.MAGIC_SLICE[0]:
                                                                 constants.MAGIC_SLICE[1]])[0]))

        self.get_target_arch(data)

        if new_header:
            self.fields = self.build_new_header(data)
        else:
            self.fields = self.build_raw_header(data)

    def get_target_arch(self, data):
        """
        Determine target architecture based on the e_machine field in the ELF
        header. e_machine is the only reliable field.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :return: None
        :rtype: None
        """
        try:
            self.e_machine = data[constants.E_MACHINE[0]:
                                  constants.E_MACHINE[1]]
            self.arch_metadata = structures.ARCH[self.e_machine]
        except KeyError as e:
            LOG.error("Unsupported architecture: " + hex(unpack('<H',
                                                                data[constants.E_MACHINE[0]:
                                                                     constants.E_MACHINE[1]])[0]))
            raise exceptions.ELFHeaderError("Error in ELF Header:"
                                            "Unsupported architecture")

    @staticmethod
    def _build_hdr(field_values, header_struct):
        """
        Populate header fields.

        :param field_values: Header field and values mapping
        :type field_values: dict
        :param header_struct: The ELF header structure definition (ELF32/ELF64)
        :type header_struct: dict
        :return: Final update of header field and values mapping
        :rtype: dict
        """
        hdr = deepcopy(header_struct)

        hdr.update({
            "e_ident": field_values["ei_mag"] + field_values["ei_class"] +
                       field_values["ei_data"] + field_values["ei_version"] +
                       field_values["ei_osabi"] + field_values["ei_abiversion"] +
                       field_values["ei_pad"],
            "e_type": field_values["e_type"],
            "e_machine": field_values["e_machine"],
            "e_version": field_values["e_version"],
            "e_entry": field_values["e_entry"],
            "e_phoff": field_values["e_phoff"],
            "e_shoff": field_values["e_shoff"],
            "e_flags": field_values["e_flags"],
            "e_ehsize": field_values["e_ehsize"],
            "e_phentsize": field_values["e_phentsize"],
            "e_phnum": field_values["e_phnum"],
            "e_shentsize": field_values["e_shentsize"],
            "e_shnum": field_values["e_shnum"],
            "e_shstrndx": field_values["e_shstrndx"],
        })

        return hdr

    def handle_overlap(self, field_values, data):
        """
        :param field_values: ELF32 header field and values mapping
        :type field_values: dict
        :param data: Input file content
        :type data: <class 'bytes'>
        :return: Updated ELF32 header field and values mapping
        :rtype: dict
        """
        pass

    def _update_elf64_values(self, field_values, data):
        """
        Update ELF64 header field and values

        :param field_values: ELF64 header field and values mapping
        :type field_values: dict
        :param data: Input file content
        :type data: <class 'bytes'>
        :return: Updated ELF64 header field and values mapping
        :rtype: dict
        """
        e_phoff = unpack("<Q", data[constants.ELF64HEADEROFFSETS.E_PHOFF[0]:
                                    constants.ELF64HEADEROFFSETS.E_PHOFF[1]])[0]
        if e_phoff < self.arch_metadata["e_phoff"]:
            LOG.info("There is likely an ELF header and program header overlap. "
                     "Lepton is not equipped to handle this scenario.")
            return None

        field_values.update({
            "e_type": data[constants.ELF64HEADEROFFSETS.E_TYPE[0]:
                           constants.ELF64HEADEROFFSETS.E_TYPE[1]],
            "e_entry": data[constants.ELF64HEADEROFFSETS.E_ENTRY[0]:
                            constants.ELF64HEADEROFFSETS.E_ENTRY[1]],
            "e_phoff": pack("<Q", self.arch_metadata["e_phoff"]),
            "e_shoff": data[constants.ELF64HEADEROFFSETS.E_SHOFF[0]:
                            constants.ELF64HEADEROFFSETS.E_SHOFF[1]],
            "e_ehsize": pack("<H", self.arch_metadata["e_ehsize"]),
            "e_phentsize": pack("<H", self.arch_metadata["e_phentsize"]),
            "e_phnum": data[constants.ELF64HEADEROFFSETS.E_PHNUM[0]:
                            constants.ELF64HEADEROFFSETS.E_PHNUM[1]],
            "e_shentsize": pack("<H", self.arch_metadata["e_shentsize"]),
            "e_shnum": data[constants.ELF64HEADEROFFSETS.E_SHNUM[0]:
                            constants.ELF64HEADEROFFSETS.E_SHNUM[1]],
            "e_shstrndx": data[constants.ELF64HEADEROFFSETS.E_SHSTRNDX[0]:
                               constants.ELF64HEADEROFFSETS.E_SHSTRNDX[1]]
        })
        return field_values

    def _update_elf32_values(self, field_values, data):
        """
        Update ELF32 header field and values

        :param field_values: ELF32 header field and values mapping
        :type field_values: dict
        :param data: Input file content
        :type data: <class 'bytes'>
        :return: Updated ELF32 header field and values mapping
        :rtype: dict
        """
        e_phoff = unpack("<I", data[constants.ELF32HEADEROFFSETS.E_PHOFF[0]:
                                    constants.ELF32HEADEROFFSETS.E_PHOFF[1]])[0]
        if e_phoff < self.arch_metadata["e_phoff"]:
            LOG.info("There is likely an ELF header and program header overlap. "
                     "Lepton is not equipped to handle this scenario.")
            return None

        field_values.update({
            "e_type": data[constants.ELF32HEADEROFFSETS.E_TYPE[0]:
                           constants.ELF32HEADEROFFSETS.E_TYPE[1]],
            "e_entry": data[constants.ELF32HEADEROFFSETS.E_ENTRY[0]:
                            constants.ELF32HEADEROFFSETS.E_ENTRY[1]],
            "e_phoff": pack("<I", self.arch_metadata["e_phoff"]),
            "e_shoff": data[constants.ELF32HEADEROFFSETS.E_SHOFF[0]:
                            constants.ELF32HEADEROFFSETS.E_SHOFF[1]],
            "e_ehsize": pack("<H", self.arch_metadata["e_ehsize"]),
            "e_phentsize": pack("<H", self.arch_metadata["e_phentsize"]),
            "e_phnum": data[constants.ELF32HEADEROFFSETS.E_PHNUM[0]:
                            constants.ELF32HEADEROFFSETS.E_PHNUM[1]],
            "e_shentsize": pack("<H", self.arch_metadata["e_shentsize"]),
            "e_shnum": data[constants.ELF32HEADEROFFSETS.E_SHNUM[0]:
                            constants.ELF32HEADEROFFSETS.E_SHNUM[1]],
            "e_shstrndx": data[constants.ELF32HEADEROFFSETS.E_SHSTRNDX[0]:
                               constants.ELF32HEADEROFFSETS.E_SHSTRNDX[1]]
        })
        return field_values

    def build_new_header(self, data):
        """
        Create a new ELF header structure and populate expected values into the
        header fields. In case of ELF header corruption, these new values should
        allow tools like pyelftools to load the input ELF file.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :return: New ELF header contents
        :rtype: dict
        """
        field_values = {
            "ei_mag": constants.ELFMAGIC,
            "ei_class": pack("B", self.arch_metadata["ei_class"]),
            "ei_data": pack("B", self.arch_metadata["ei_data"]),
            "ei_version": pack("B", self.arch_metadata["ei_version"]),
            "ei_osabi": pack("B", self.arch_metadata["ei_osabi"]),
            "ei_abiversion": pack("B", self.arch_metadata["ei_abiversion"]),
            "ei_pad": self.arch_metadata["padding"],
            "e_machine": self.e_machine,
            "e_version": pack("<I", self.arch_metadata["e_version"]),
            "e_flags": pack("<I", 0)
        }

        if self.arch_metadata["ei_class"] == 1:
            field_values = self._update_elf32_values(field_values, data)
            if field_values is None:
                return
            return self._build_hdr(field_values, deepcopy(structures.ELF32HEADER))
        else:
            field_values = self._update_elf64_values(field_values, data)
            if field_values is None:
                return
            return self._build_hdr(field_values, deepcopy(structures.ELF64HEADER))

    def build_raw_header(self, data):
        """
        Create a header structure from the input file itself. Essentially, cloning
        it.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :return: Cloned ELF header contents
        :rtype: dict
        """
        if self.arch_metadata["ei_class"] == 1:
            field_values = {
                "ei_mag": constants.ELFMAGIC,
                "ei_class": data[constants.ELF32HEADEROFFSETS.EI_CLASS[0]:
                                 constants.ELF32HEADEROFFSETS.EI_CLASS[1]],
                "ei_data": data[constants.ELF32HEADEROFFSETS.EI_DATA[0]:
                                constants.ELF32HEADEROFFSETS.EI_DATA[1]],
                "ei_version": data[constants.ELF32HEADEROFFSETS.EI_VERSION[0]:
                                   constants.ELF32HEADEROFFSETS.EI_VERSION[1]],
                "ei_osabi": data[constants.ELF32HEADEROFFSETS.EI_OSABI[0]:
                                 constants.ELF32HEADEROFFSETS.EI_OSABI[1]],
                "ei_abiversion": data[constants.ELF32HEADEROFFSETS.EI_ABIVERSION[0]:
                                      constants.ELF32HEADEROFFSETS.EI_ABIVERSION[1]],
                "ei_pad": data[constants.ELF32HEADEROFFSETS.EI_PAD[0]:
                               constants.ELF32HEADEROFFSETS.EI_PAD[1]],
                "e_type": data[constants.ELF32HEADEROFFSETS.E_TYPE[0]:
                               constants.ELF32HEADEROFFSETS.E_TYPE[1]],
                "e_machine": data[constants.ELF32HEADEROFFSETS.E_MACHINE[0]:
                                  constants.ELF32HEADEROFFSETS.E_MACHINE[1]],
                "e_version": data[constants.ELF32HEADEROFFSETS.E_VERSION[0]:
                                  constants.ELF32HEADEROFFSETS.E_VERSION[1]],
                "e_entry": data[constants.ELF32HEADEROFFSETS.E_ENTRY[0]:
                                constants.ELF32HEADEROFFSETS.E_ENTRY[1]],
                "e_phoff": data[constants.ELF32HEADEROFFSETS.E_PHOFF[0]:
                                constants.ELF32HEADEROFFSETS.E_PHOFF[1]],
                "e_shoff": data[constants.ELF32HEADEROFFSETS.E_SHOFF[0]:
                                constants.ELF32HEADEROFFSETS.E_SHOFF[1]],
                "e_flags": data[constants.ELF32HEADEROFFSETS.E_FLAGS[0]:
                                constants.ELF32HEADEROFFSETS.E_FLAGS[1]],
                "e_ehsize": data[constants.ELF32HEADEROFFSETS.E_EHSIZE[0]:
                                 constants.ELF32HEADEROFFSETS.E_EHSIZE[1]],
                "e_phentsize": data[constants.ELF32HEADEROFFSETS.E_PHENTSIZE[0]:
                                    constants.ELF32HEADEROFFSETS.E_PHENTSIZE[1]],
                "e_phnum": data[constants.ELF32HEADEROFFSETS.E_PHNUM[0]:
                                constants.ELF32HEADEROFFSETS.E_PHNUM[1]],
                "e_shentsize": data[constants.ELF32HEADEROFFSETS.E_SHENTSIZE[0]:
                                    constants.ELF32HEADEROFFSETS.E_SHENTSIZE[1]],
                "e_shnum": data[constants.ELF32HEADEROFFSETS.E_SHNUM[0]:
                                constants.ELF32HEADEROFFSETS.E_SHNUM[1]],
                "e_shstrndx": data[constants.ELF32HEADEROFFSETS.E_SHSTRNDX[0]:
                                   constants.ELF32HEADEROFFSETS.E_SHSTRNDX[1]],
            }
            return self._build_hdr(field_values, deepcopy(structures.ELF32HEADER))
        else:
            field_values = {
                "ei_mag": constants.ELFMAGIC,
                "ei_class": data[constants.ELF64HEADEROFFSETS.EI_CLASS[0]:
                                 constants.ELF64HEADEROFFSETS.EI_CLASS[1]],
                "ei_data": data[constants.ELF64HEADEROFFSETS.EI_DATA[0]:
                                constants.ELF64HEADEROFFSETS.EI_DATA[1]],
                "ei_version": data[constants.ELF64HEADEROFFSETS.EI_VERSION[0]:
                                   constants.ELF64HEADEROFFSETS.EI_VERSION[1]],
                "ei_osabi": data[constants.ELF64HEADEROFFSETS.EI_OSABI[0]:
                                 constants.ELF64HEADEROFFSETS.EI_OSABI[1]],
                "ei_abiversion": data[constants.ELF64HEADEROFFSETS.EI_ABIVERSION[0]:
                                      constants.ELF64HEADEROFFSETS.EI_ABIVERSION[1]],
                "ei_pad": data[constants.ELF64HEADEROFFSETS.EI_PAD[0]:
                               constants.ELF64HEADEROFFSETS.EI_PAD[1]],
                "e_type": data[constants.ELF64HEADEROFFSETS.E_TYPE[0]:
                               constants.ELF64HEADEROFFSETS.E_TYPE[1]],
                "e_machine": data[constants.ELF64HEADEROFFSETS.E_MACHINE[0]:
                                  constants.ELF64HEADEROFFSETS.E_MACHINE[1]],
                "e_version": data[constants.ELF64HEADEROFFSETS.E_VERSION[0]:
                                  constants.ELF64HEADEROFFSETS.E_VERSION[1]],
                "e_entry": data[constants.ELF64HEADEROFFSETS.E_ENTRY[0]:
                                constants.ELF64HEADEROFFSETS.E_ENTRY[1]],
                "e_phoff": data[constants.ELF64HEADEROFFSETS.E_PHOFF[0]:
                                constants.ELF64HEADEROFFSETS.E_PHOFF[1]],
                "e_shoff": data[constants.ELF64HEADEROFFSETS.E_SHOFF[0]:
                                constants.ELF64HEADEROFFSETS.E_SHOFF[1]],
                "e_flags": data[constants.ELF64HEADEROFFSETS.E_FLAGS[0]:
                                constants.ELF64HEADEROFFSETS.E_FLAGS[1]],
                "e_ehsize": data[constants.ELF64HEADEROFFSETS.E_EHSIZE[0]:
                                 constants.ELF64HEADEROFFSETS.E_EHSIZE[1]],
                "e_phentsize": data[constants.ELF64HEADEROFFSETS.E_PHENTSIZE[0]:
                                    constants.ELF64HEADEROFFSETS.E_PHENTSIZE[1]],
                "e_phnum": data[constants.ELF64HEADEROFFSETS.E_PHNUM[0]:
                                constants.ELF64HEADEROFFSETS.E_PHNUM[1]],
                "e_shentsize": data[constants.ELF64HEADEROFFSETS.E_SHENTSIZE[0]:
                                    constants.ELF64HEADEROFFSETS.E_SHENTSIZE[1]],
                "e_shnum": data[constants.ELF64HEADEROFFSETS.E_SHNUM[0]:
                                constants.ELF64HEADEROFFSETS.E_SHNUM[1]],
                "e_shstrndx": data[constants.ELF64HEADEROFFSETS.E_SHSTRNDX[0]:
                                   constants.ELF64HEADEROFFSETS.E_SHSTRNDX[1]],
            }
            return self._build_hdr(field_values, deepcopy(structures.ELF64HEADER))

    def to_bytes(self):
        """
        Returns ELF header contents in the form of a sequence of bytes
        :return: ELF header contents
        :rtype: <class 'bytes'>
        """
        return self.fields["e_ident"] + self.fields["e_type"] + \
               self.fields["e_machine"] + self.fields["e_version"] + \
               self.fields["e_entry"] + self.fields["e_phoff"] + \
               self.fields["e_shoff"] + self.fields["e_flags"] + \
               self.fields["e_ehsize"] + self.fields["e_phentsize"] + \
               self.fields["e_phnum"] + self.fields["e_shentsize"] + \
               self.fields["e_shnum"] + self.fields["e_shstrndx"]



import os
import logging
from copy import deepcopy
from struct import unpack

import lepton.utils.constants as constants
import lepton.utils.exceptions as exceptions
import lepton.utils.structures as structures

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
LOG = logging.getLogger(__name__)


class ELFProgramHeaderTable:
    def __init__(self, data, elfheader):
        """
        :param data: Input file contents
        :type data: <class 'bytes'>
        :param elfheader: Lepton-constructed ELF header contents
        :type elfheader: dict
        """
        self.entries = self.build_program_header_table(data, elfheader)

    @staticmethod
    def _update_elf32_values(data, pheader_struct, phdr_num):
        """
        Update ELF32 program header table entry fields with values.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param pheader_struct: ELF32 program header table entry definition.
        :type pheader_struct: dict
        :param phdr_num: Program header table entry number. 0-indexed.
        :type phdr_num: int
        :return: Constructed ELF32 program header table entry
        :rtype: dict
        """
        # Calculate offset of p_hdr
        # p_hdr_offset = e_phoff
        e_phoff = unpack('<I', data[constants.ELF32HEADEROFFSETS.E_PHOFF[0]:
                                    constants.ELF32HEADEROFFSETS.E_PHOFF[1]])[0]
        e_phentsize = unpack('<H', data[constants.ELF32HEADEROFFSETS.E_PHENTSIZE[0]:
                                        constants.ELF32HEADEROFFSETS.E_PHENTSIZE[1]])[0]
        phdr = deepcopy(pheader_struct)
        # p_hdr_i_offset = e_phoff + e_phentsize * phdr_num
        phdr_entry_offset = e_phoff + e_phentsize * phdr_num

        # The program header fields must be accurate in the binary since these are needed
        # by the ELF loader.
        phdr["p_type"] = data[phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_TYPE[0]:
                              phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_TYPE[1]]
        phdr["p_offset"] = data[phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_OFFSET[0]:
                                phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_OFFSET[1]]
        phdr["p_vaddr"] = data[phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_VADDR[0]:
                               phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_VADDR[1]]
        phdr["p_paddr"] = data[phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_PADDR[0]:
                               phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_PADDR[1]]
        phdr["p_filesz"] = data[phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_FILESZ[0]:
                                phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_FILESZ[1]]
        phdr["p_memsz"] = data[phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_MEMSZ[0]:
                               phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_MEMSZ[1]]
        phdr["p_flags"] = data[phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_FLAGS[0]:
                               phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_FLAGS[1]]
        phdr["p_align"] = data[phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_ALIGN[0]:
                               phdr_entry_offset + constants.ELF32PROGRAMHEADEROFFSETS.P_ALIGN[1]]

        return phdr

    @staticmethod
    def _update_elf64_values(data, pheader_struct, phdr_num):
        """
        Update ELF64 program header table entry fields with values.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param pheader_struct: ELF64 program header table entry definition.
        :type pheader_struct: dict
        :param phdr_num: Program header table entry number. 0-indexed.
        :type phdr_num: int
        :return: Constructed ELF64 program header table entry
        :rtype: dict
        """
        # Calculate offset of p_hdr
        # p_hdr_offset = e_phoff
        e_phoff = unpack('<Q', data[constants.ELF64HEADEROFFSETS.E_PHOFF[0]:
                                    constants.ELF64HEADEROFFSETS.E_PHOFF[1]])[0]
        e_phentsize = unpack('<H', data[constants.ELF64HEADEROFFSETS.E_PHENTSIZE[0]:
                                        constants.ELF64HEADEROFFSETS.E_PHENTSIZE[1]])[0]
        # p_hdr_i_offset = e_phoff + e_phentsize * phdr_num
        phdr_entry_offset = e_phoff + e_phentsize * phdr_num

        phdr = deepcopy(pheader_struct)

        # The program header fields must be accurate in the binary since these are needed
        # by the ELF loader.
        phdr["p_type"] = data[phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_TYPE[0]:
                              phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_TYPE[1]]
        phdr["p_offset"] = data[phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_OFFSET[0]:
                                phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_OFFSET[1]]
        phdr["p_vaddr"] = data[phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_VADDR[0]:
                               phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_VADDR[1]]
        phdr["p_paddr"] = data[phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_PADDR[0]:
                               phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_PADDR[1]]
        phdr["p_filesz"] = data[phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_FILESZ[0]:
                                phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_FILESZ[1]]
        phdr["p_memsz"] = data[phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_MEMSZ[0]:
                               phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_MEMSZ[1]]
        phdr["p_flags"] = data[phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_FLAGS[0]:
                               phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_FLAGS[1]]
        phdr["p_align"] = data[phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_ALIGN[0]:
                               phdr_entry_offset + constants.ELF64PROGRAMHEADEROFFSETS.P_ALIGN[1]]

        return phdr

    def _build_phdr(self, data, pheader_struct, phdr_num):
        """
        Router function. Calls relevant functions to build program header table
        for an ELF32/ELF64 file.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param pheader_struct: ELF32/64 program header table entry definition.
        :type pheader_struct: dict
        :param phdr_num: Program header table entry number. 0-indexed.
        :type phdr_num: int
        :return: Constructed ELF program header table
        :rtype: list of dict
        """
        if pheader_struct == structures.ELF32PROGRAMHEADER:
            return self._update_elf32_values(data, pheader_struct, phdr_num)
        elif pheader_struct == structures.ELF64PROGRAMHEADER:
            return self._update_elf64_values(data, pheader_struct, phdr_num)
        else:
            raise exceptions.ELFProgramHeaderTableError("Invalid Program"
                                                        "Header Structure.")

    def build_program_header_table(self, data, elfheader):
        """
        Builds the program header table of the input ELF file.

        :param data: Input file contents
        :type data: <class 'bytes'>
        :param elfheader: Lepton-constructed ELF header contents
        :type elfheader: dict
        :return: Constructed ELF program header table
        :rtype: list of dict
        """
        phdr_table = []
        phnum = unpack("<H", elfheader["e_phnum"])[0]
        ei_class = elfheader["e_ident"][4]

        for phdr_num in range(phnum):
            if ei_class == 1:
                phdr = self._build_phdr(data, structures.ELF32PROGRAMHEADER,
                                        phdr_num)
            else:
                phdr = self._build_phdr(data, structures.ELF64PROGRAMHEADER,
                                        phdr_num)

            phdr_table.append(phdr)

        return phdr_table

    def to_bytes(self):
        """
        Returns ELF program header contents in the form of a sequence of bytes
        :return: ELF program header contents
        :rtype: <class 'bytes'>
        """
        phdr_table_bytes = bytes()
        if len(self.entries[0]["p_vaddr"]) == 8:
            # ELF64
            for phdr in self.entries:
                header_bytes = phdr["p_type"] + phdr["p_flags"] + \
                               phdr["p_offset"] + phdr["p_vaddr"] + \
                               phdr["p_paddr"] + phdr["p_filesz"] + \
                               phdr["p_memsz"] + phdr["p_align"]
                phdr_table_bytes += header_bytes
        else:
            for phdr in self.entries:
                header_bytes = phdr["p_type"] + phdr["p_offset"] + \
                               phdr["p_vaddr"] + phdr["p_paddr"] + \
                               phdr["p_filesz"] + phdr["p_memsz"] + \
                               phdr["p_flags"] + phdr["p_align"]
                phdr_table_bytes += header_bytes
        return phdr_table_bytes



import os
import logging

from lepton.utils.exceptions import *
from lepton.elfheader import ELFHeader
from lepton.elfprogramheader import ELFProgramHeaderTable
from lepton.elfsectionheader import ELFSectionHeaderTable

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
LOG = logging.getLogger(__name__)


class ELFFile:
    def __init__(self, file_handle, new_header=False):
        """
        :param file_handle: File handle to file
        :type file_handle: <class '_io.BufferedReader'>
        :param new_header: Flag to decide whether to build a new header with
                           standard values or clone the header of the input file.
        :type new_header: bool
        """
        self.new_header = new_header

        try:
            self.data = file_handle.read()
        except PermissionError as err:
            LOG.error(f"Permission error: {err}")
            return

        try:
            self.elfheader = ELFHeader(self.data, new_header)
        except ELFHeaderError as err:
            LOG.error(f"ELF header error: {err}")
            return

        if self.elfheader.fields is None:
            LOG.error("ELF header was not constructed. Exiting.")
            return

        try:
            self.elfprogheader = ELFProgramHeaderTable(
                                     self.data,
                                     self.elfheader.fields
                                 )
        except ELFProgramHeaderTableError as err:
            LOG.error(f"Program header table error: {err}")
            return

        try:
            self.elfsectionheader = ELFSectionHeaderTable(
                                        self.data,
                                        self.elfheader.fields,
                                        new_header
                                    )
        except ELFSectionHeaderTableError as err:
            LOG.error(f"Section header table error: {err}")
            return

    def reconstruct_file(self):
        """
        This function reconstructs the ELF file with Lepton-constructed headers
        and non-header content directly from the input file.
        """
        # Initial new file with ELF header contents
        new_data = self.elfheader.to_bytes()

        # Add any bytes between ELF header and program headers table
        e_phoff = self.elfheader.fields["e_phoff"]
        gap = int.from_bytes(e_phoff, byteorder="little") - len(new_data)
        if gap:
            new_data += self.data[len(new_data): len(new_data) + gap]

        # Add program headers table to new file
        new_data += self.elfprogheader.to_bytes()

        # Add any bytes between bytes add so far and section headers table
        e_shoff = self.elfheader.fields["e_shoff"]
        gap = int.from_bytes(e_shoff, byteorder="little") - len(new_data)
        if gap:
            new_data += self.data[len(new_data): len(new_data) + gap]

        # Add section headers table to new file
        new_data += self.elfsectionheader.to_bytes()

        # Add rest of the data
        gap = len(self.data) - len(new_data)
        if gap:
            new_data += self.data[len(new_data): len(new_data) + gap]

        return new_data



ELFMAGIC = b"\x7fELF"
MAGIC_SLICE = (0, 4)
E_MACHINE = (0x12, 0x14)


class ELF32HEADEROFFSETS:
    MAGIC_SLICE = (0, 4)
    EI_CLASS = (4, 5)
    EI_DATA = (5, 6)
    EI_VERSION = (6, 7)
    EI_OSABI = (7, 8)
    EI_ABIVERSION = (8, 9)
    EI_PAD = (9, 0x10)
    E_TYPE = (0x10, 0x12)
    E_MACHINE = (0x12, 0x14)
    E_VERSION = (0x14, 0x18)
    E_ENTRY = (0x18, 0x1C)
    E_PHOFF = (0x1C, 0x20)
    E_SHOFF = (0x20, 0x24)
    E_FLAGS = (0x24, 0x28)
    E_EHSIZE = (0x28, 0x2A)
    E_PHENTSIZE = (0x2A, 0x2C)
    E_PHNUM = (0x2C, 0x2E)
    E_SHENTSIZE = (0x2E, 0x30)
    E_SHNUM = (0x30, 0x32)
    E_SHSTRNDX = (0x32, 0x34)


class ELF64HEADEROFFSETS:
    MAGIC_SLICE = (0, 4)
    EI_CLASS = (4, 5)
    EI_DATA = (5, 6)
    EI_VERSION = (6, 7)
    EI_OSABI = (7, 8)
    EI_ABIVERSION = (8, 9)
    EI_PAD = (9, 0x10)
    E_TYPE = (0x10, 0x12)
    E_MACHINE = (0x12, 0x14)
    E_VERSION = (0x14, 0x18)
    E_ENTRY = (0x18, 0x20)
    E_PHOFF = (0x20, 0x28)
    E_SHOFF = (0x28, 0x30)
    E_FLAGS = (0x30, 0x34)
    E_EHSIZE = (0x34, 0x36)
    E_PHENTSIZE = (0x36, 0x38)
    E_PHNUM = (0x38, 0x3A)
    E_SHENTSIZE = (0x3A, 0x3C)
    E_SHNUM = (0x3C, 0x3E)
    E_SHSTRNDX = (0x3E, 0x40)


class ELF32PROGRAMHEADEROFFSETS:
    P_TYPE = (0, 4)
    P_OFFSET = (4, 8)
    P_VADDR = (8, 12)
    P_PADDR = (12, 16)
    P_FILESZ = (16, 20)
    P_MEMSZ = (20, 24)
    P_FLAGS = (24, 28)
    P_ALIGN = (28, 32)


class ELF64PROGRAMHEADEROFFSETS:
    P_TYPE = (0, 4)
    P_FLAGS = (4, 8)
    P_OFFSET = (8, 16)
    P_VADDR = (16, 24)
    P_PADDR = (24, 32)
    P_FILESZ = (32, 40)
    P_MEMSZ = (40, 48)
    P_ALIGN = (48, 56)


class ELF32SECTIONHEADEROFFSETS:
    SH_NAME = (0, 4)
    SH_TYPE = (4, 8)
    SH_FLAGS = (8, 12)
    SH_ADDR = (12, 16)
    SH_OFFSET = (16, 20)
    SH_SIZE = (20, 24)
    SH_LINK = (24, 28)
    SH_INFO = (28, 32)
    SH_ADDRALIGN = (32, 36)
    SH_ENTSIZE = (36, 40)


class ELF64SECTIONHEADEROFFSETS:
    SH_NAME = (0, 4)
    SH_TYPE = (4, 8)
    SH_FLAGS = (8, 16)
    SH_ADDR = (16, 24)
    SH_OFFSET = (24, 32)
    SH_SIZE = (32, 40)
    SH_LINK = (40, 44)
    SH_INFO = (44, 48)
    SH_ADDRALIGN = (48, 56)
    SH_ENTSIZE = (56, 64)



from setuptools import setup, find_packages


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name="elflepton",
    version="1.0.2",
    license="MIT",
    description="Parse an ELF binary with corrupted ELF headers.",
    long_description=readme(),
    long_description_content_type='text/markdown',
    author="Nikhil Ashok Hegde",
    author_email="nikhilhegde20@gmail.com",
    url="https://github.com/nikhilh-20/lepton",
    packages=find_packages()
)
